#include "LCDTouch.h"
#include "Arduino.h"
#include "UFont.h"
/**********************************************/
#if defined(__AVR__)
	#include <avr/pgmspace.h>
#endif

#if defined(__AVR_ATmega1280__) || defined(__AVR_ATmega2560__)
	#include "HW_ATmega1280.h"
		
#elif defined(__AVR_ATmega328P__)

	#include "HW_ATmega328P.h"
	#define MainChip    419	
#elif defined(__AVR_ATmega32U4__)
	#include "HW_ATmega32U4.h"
	#define MainChip    419	
#elif defined(__AVR_ATmega168__)
	#error "ATmega168 MCUs are not supported because they have too little flash memory!"
#else 
	#error "Unsupported AVR MCU!"
#endif

unsigned int gColor;
extern fontdatatype PSmallFont[1144];
extern fontdatatype PBigFont[3044];

LCDTouch Touchtest;

extern byte orient;

void LCDTouch::SetXY(unsigned int x0,unsigned int y0,unsigned int x1,unsigned int y1)
{
  unsigned int i;
  if(display_model == LCD2_4)
  {
	swap(word, x0, y0);
	swap(word, x1, y1)
	y0=disp_y_size-y0;
	y1=disp_y_size-y1;
	swap(word, y0, y1)
	
	Write_Command_Data(0x0046,(x1 << 8)| x0);
	Write_Command_Data(0x0047,y1);
	Write_Command_Data(0x0048,y0);
	Write_Command_Data(0x0020,x0);
	Write_Command_Data(0x0021,y0);
	Write_Command (0x0022);
  
  }
  else if(display_model == LCD2_8)
  {
	swap(word, x0, y0);
	swap(word, x1, y1)
	y0=disp_y_size-y0;
	y1=disp_y_size-y1;
	swap(word, y0, y1)
	
	Write_Command_Data(WINDOW_XADDR_START,x0);
	Write_Command_Data(WINDOW_YADDR_START,y0);
	Write_Command_Data(WINDOW_XADDR_END,x1);
	Write_Command_Data(WINDOW_YADDR_END,y1);
	Write_Command_Data(GRAM_XADDR,x0);
	Write_Command_Data(GRAM_YADDR,y0);
	Write_Command (0x0022);
  }
  else
  {
	swap(word, x0, y0);
	swap(word, x1, y1)
	y0=disp_y_size-y0;
	y1=disp_y_size-y1;
	swap(word, y0, y1)
	
	Write_Command_Data(0x0044,(x1<<8)+x0);
	Write_Command_Data(0x0045,y0);
	Write_Command_Data(0x0046,y1);
	Write_Command_Data(0x004e,x0);
	Write_Command_Data(0x004f,y0);
	Write_Command (0x0022);
  }
}
void LCDTouch::Pant(unsigned int color)
{
	int i,j;
	SetXY(0,0,320,240);

    for(i=0;i<320;i++)
	{
	  for (j=0;j<240;j++)
	   	{
			Write_Data(color);
	    }

	}		
}
void LCDTouch::fillScr(unsigned int color)
{
if(display_model == LCD2_4 || display_model == LCD2_8)
  {
	int i,j;
	SetXY(0,0,320,240);

    for(i=0;i<320;i++)
	 {
	  for (j=0;j<240;j++)
	   	{
         Write_Data(color);
	    }
	 }
	}
else
{
	int i,j;
	SetXY(0,0,319,239);

    for(i=0;i<320;i++)
	{
	  for (j=0;j<240;j++)
	   	{
         Write_Data(color);
	    }

	}	
}
}
void LCDTouch::fillScr(byte r, byte g, byte b)
{
	word color = ((r&248)<<8 | (g&252)<<3 | (b&248)>>3);
	fillScr(color);
}
void LCDTouch::clrScr()
{
	unsigned int i,j;
	SetXY(0,0,319,239);
	for(i=0;i<240;i++)
	{
	    for(j=0;j<320;j++)
		{    
          	Write_Data(0x0000);
		}
	}
	
}

void LCDTouch::setPrecision(byte precision)
{
	switch (precision)
	{
		case PREC_LOW:
			prec=1;
			break;
		case PREC_MEDIUM:
			prec=10;
			break;
		case PREC_HI:
			prec=25;
			break;
		case PREC_EXTREME:
			prec=100;
			break;
		default:
			prec=10;
			break;
	}
}

void LCDTouch::LcdOff()
{
	cbi(P_CS, B_CS);
	Write_Data(0x28);
	sbi(P_CS, B_CS);
}

void LCDTouch::LcdOn()
{
	cbi(P_CS, B_CS);
	Write_Data(0x29);
	sbi(P_CS, B_CS);
}
void LCDTouch::clrXY()
{
	if (orient==PORTRAIT)
		{
			SetXY(0,0,disp_x_size,disp_y_size);
		}
	else
		{
			SetXY(0,0,disp_y_size,disp_x_size);
		}
}
void LCDTouch::setColor(word color)
{
	fch=byte(color>>8);
	fcl=byte(color & 0xFF);
}
void LCDTouch::setColor(byte r, byte g, byte b)
{
	fch=((r&248)|g>>5);
	fcl=((g&28)<<3|b>>3);
}
void LCDTouch::setBackColor(word color)
{
	bch=byte(color>>8);
	bcl=byte(color & 0xFF);
}
void LCDTouch::setBackColor(byte r, byte g, byte b)
{
	bch=((r&248)|g>>5);
	bcl=((g&28)<<3|b>>3);
}
void LCDTouch::drawRect(int x1, int y1, int x2, int y2)
{
	int tmp;

	if (x1>x2)
	{
		swap(int, x1, x2);
	}
	if (y1>y2)
	{
		swap(int, y1, y2);
	}
	drawHLine(x1, y1, x2-x1);
	drawHLine(x1, y2, x2-x1);
	drawVLine(x1, y1, y2-y1);
	drawVLine(x2, y1, y2-y1);
}
void LCDTouch::drawHLine(int x, int y, int l)
{
	int i,j;
	if (l<0)
	{
		l = -l;
		y -= l;
	}
	
	SetXY(x,y,x+l,y);

//    for(i=0;i<320;i++)
//	{
	  for (j=x;j<x+l;j++)
	   	{
			Write_Data((fch<<8)|fcl);
	    }

//	}
	clrXY();
}

void LCDTouch::drawVLine(int x, int y, int l)
{
	int i,j;
	if (l<0)
	{
		l = -l;
		y -= l;
	}
	
	SetXY(x,y,x,y+l);

//    for(i=0;i<320;i++)
//	{
	  for (j=y;j<y+l;j++)
	   	{
			Write_Data((fch<<8)|fcl);
	    }

//	}
	clrXY();
}
void LCDTouch::drawLine(int x1, int y1, int x2, int y2)
{
	if (y1==y2)
		drawHLine(x1, y1, x2-x1);
	else if (x1==x2)
		drawVLine(x1, y1, y2-y1);
	else
	{
		unsigned int	dx = (x2 > x1 ? x2 - x1 : x1 - x2);
		short			xstep =  x2 > x1 ? 1 : -1;
		unsigned int	dy = (y2 > y1 ? y2 - y1 : y1 - y2);
		short			ystep =  y2 > y1 ? 1 : -1;
		int				col = x1, row = y1;

		if (dx < dy)
		{
			int t = - (dy >> 1);
			while (true)
			{
				SetXY (col, row, col, row);
				LCD_Write_DATA (fch, fcl);
				if (row == y2)
					return;
				row += ystep;
				t += dx;
				if (t >= 0)
				{
					col += xstep;
					t   -= dy;
				}
			} 
		}
		else
		{
			int t = - (dx >> 1);
			while (true)
			{
				SetXY (col, row, col, row);
				LCD_Write_DATA (fch, fcl);
				if (col == x2)
					return;
				col += xstep;
				t += dy;
				if (t >= 0)
				{
					row += ystep;
					t   -= dx;
				}
			} 
		}
	}
	clrXY();
}
void LCDTouch::drawRoundRect(int x1, int y1, int x2, int y2)
{
	int tmp;

	if (x1>x2)
	{
		swap(int, x1, x2);
	}
	if (y1>y2)
	{
		swap(int, y1, y2);
	}
	if ((x2-x1)>4 && (y2-y1)>4)
	{
		drawPixel(x1+1,y1+1);
		drawPixel(x2-1,y1+1);
		drawPixel(x1+1,y2-1);
		drawPixel(x2-1,y2-1);
		drawHLine(x1+2, y1, x2-x1-4);
		drawHLine(x1+2, y2, x2-x1-4);
		drawVLine(x1, y1+2, y2-y1-4);
		drawVLine(x2, y1+2, y2-y1-4);
	}
}

void LCDTouch::drawPixel(int x, int y)
{
//	cbi(P_CS, B_CS);
	SetXY(x, y, x, y);
	setPixel((fch<<8)|fcl);
//	sbi(P_CS, B_CS);
//	clrXY();
}
void LCDTouch::setPixel(word color)
{
	Write_Data(color);	
}
void LCDTouch::fillRect(int x1, int y1, int x2, int y2)
{
	int tmp;

	if (x1>x2)
	{
		swap(int, x1, x2);
	}
	if (y1>y2)
	{
		swap(int, y1, y2);
	}
	SetXY(x1, y1, x2, y2);
	Write_Data((fch<<8)|fcl);	
	Write_Data(((long(x2-x1)+1)*(long(y2-y1)+1)));

	if (orient==PORTRAIT)
	{
		for (int i=0; i<((y2-y1)/2)+1; i++)
		{
			drawHLine(x1, y1+i, x2-x1);
			drawHLine(x1, y2-i, x2-x1);
		}
	}
	else
	{
		for (int i=0; i<((x2-x1)/2)+1; i++)
		{
			drawVLine(x1+i, y1, y2-y1);
			drawVLine(x2-i, y1, y2-y1);
		}
	}
	
	
}
void LCDTouch::fillRoundRect(int x1, int y1, int x2, int y2)
{
	int tmp;

	if (x1>x2)
	{
		swap(int, x1, x2);
	}
	if (y1>y2)
	{
		swap(int, y1, y2);
	}

	if ((x2-x1)>4 && (y2-y1)>4)
	{
		for (int i=0; i<((y2-y1)/2)+1; i++)
		{
			switch(i)
			{
			case 0:
				drawHLine(x1+2, y1+i, x2-x1-4);
				drawHLine(x1+2, y2-i, x2-x1-4);
				break;
			case 1:
				drawHLine(x1+1, y1+i, x2-x1-2);
				drawHLine(x1+1, y2-i, x2-x1-2);
				break;
			default:
				drawHLine(x1, y1+i, x2-x1);
				drawHLine(x1, y2-i, x2-x1);
			}
		}
	}
}
void LCDTouch::drawCircle(int x, int y, int radius)
{
	int f = 1 - radius;
	int ddF_x = 1;
	int ddF_y = -2 * radius;
	int x1 = 0;
	int y1 = radius;

	SetXY(x, y + radius, x, y + radius);
	Write_Data(gColor);
	SetXY(x, y - radius, x, y - radius);
	Write_Data(gColor);
	SetXY(x + radius, y, x + radius, y);
	Write_Data(gColor);
	SetXY(x - radius, y, x - radius, y);
	Write_Data(gColor);
 
	while(x1 < y1)
	{
		if(f >= 0) 
		{
			y1--;
			ddF_y += 2;
			f += ddF_y;
		}
		x1++;
		ddF_x += 2;
		f += ddF_x;    
		SetXY(x + x1, y + y1, x + x1, y + y1);
		Write_Data(gColor);
		SetXY(x - x1, y + y1, x - x1, y + y1);
		Write_Data(gColor);
		SetXY(x + x1, y - y1, x + x1, y - y1);
		Write_Data(gColor);
		SetXY(x - x1, y - y1, x - x1, y - y1);
		Write_Data(gColor);
		SetXY(x + y1, y + x1, x + y1, y + x1);
		Write_Data(gColor);
		SetXY(x - y1, y + x1, x - y1, y + x1);
		Write_Data(gColor);
		SetXY(x + y1, y - x1, x + y1, y - x1);
		Write_Data(gColor);
		SetXY(x - y1, y - x1, x - y1, y - x1);
		Write_Data(gColor);
	}
	clrXY();
}
void LCDTouch::fillCircle(int x, int y, int radius)
{
	for(int y1=-radius; y1<=0; y1++) 
		for(int x1=-radius; x1<=0; x1++)
			if(x1*x1+y1*y1 <= radius*radius) 
			{
				drawHLine(x+x1, y+y1, 2*(-x1));
				drawHLine(x+x1, y-y1, 2*(-x1));
				break;
			}
}
void LCDTouch::setFont(uint8_t mxsize,uint8_t mysize,uint8_t moffset)
{
	cfont.x_size=mxsize;
	cfont.y_size=mysize;
	cfont.offset=moffset;
}
void LCDTouch::printChar(byte c, int x, int y)
{		
	byte i,ch;
	word j;
	word temp; 
	if(cfont.x_size == 16)
	{
		cfont.font = &PBigFont[0];
	}
	else if(cfont.x_size == 32)
	{
		cfont.font = &PSevenSegNumFont[0];
	}
	else
	{
		cfont.font = &PSmallFont[0];
	}
	if (orient==PORTRAIT)
	{
		SetXY(x,y,x+cfont.x_size-1,y+cfont.y_size-1);
	  
		temp=((c-cfont.offset)*((cfont.x_size/8)*cfont.y_size))+4;
		for(j=0;j<((cfont.x_size/8)*cfont.y_size);j++)
		{
			ch=pgm_read_byte(&cfont.font[temp]);
			for(i=0;i<8;i++)
			{   
				if((ch&(1<<(7-i)))!=0)   
				{
					setPixel((fch<<8)|fcl);
				} 
				else
				{
					setPixel((bch<<8)|bcl);
				}   
			}
			temp++;
		}
	}
	else
	{
		temp=((c-cfont.offset)*((cfont.x_size/8)*cfont.y_size))+4;

		for(j=0;j<((cfont.x_size/8)*cfont.y_size);j+=(cfont.x_size/8))
		{
			SetXY(x,y+(j/(cfont.x_size/8)),x+cfont.x_size-1,y+(j/(cfont.x_size/8)));
			for (int zz=(cfont.x_size/8)-1; zz>=0; zz--)
			{
				ch=pgm_read_byte(&cfont.font[temp+zz]);
				for(i=0;i<8;i++)
				{   
					if((ch&(1<<i))!=0)   
					{
						setPixel((fch<<8)|fcl);
					} 
					else
					{
						setPixel((bch<<8)|bcl);
					}   
				}
			}
			temp+=(cfont.x_size/8);
		}
	}
	clrXY();
}
void LCDTouch::rotateChar(byte c, int x, int y, int pos, int deg)
{
	byte i,j,ch;
	word temp; 
	int newx,newy;
	double radian;
	radian=deg*0.0175;  

	cfont.font = &PSmallFont[0];
	
	temp=((c-cfont.offset)*((cfont.x_size/8)*cfont.y_size))+4;
	for(j=0;j<cfont.y_size;j++) 
	{
		for (int zz=0; zz<(cfont.x_size/8); zz++)
		{
			ch=pgm_read_byte(&cfont.font[temp+zz]); 
			for(i=0;i<8;i++)
			{   
				newx=x+(((i+(zz*8)+(pos*cfont.x_size))*cos(radian))-((j)*sin(radian)));
				newy=y+(((j)*cos(radian))+((i+(zz*8)+(pos*cfont.x_size))*sin(radian)));

				SetXY(newx,newy,newx+1,newy+1);
				
				if((ch&(1<<(7-i)))!=0)   
				{
					setPixel((fch<<8)|fcl);
				} 
				else  
				{
					setPixel((bch<<8)|bcl);
				}   
			}
		}
		temp+=(cfont.x_size/8);
	}
	clrXY();
}
void LCDTouch::print(char *st, int x, int y, int deg)
{
	int stl, i;

	stl = strlen(st);

	if (orient==PORTRAIT)
	{
	if (x==RIGHT)
		x=(disp_x_size+1)-(stl*cfont.x_size);
	if (x==CENTER)
		x=((disp_x_size+1)-(stl*cfont.x_size))/2;
	}
	else
	{
	if (x==RIGHT)
		x=(disp_y_size+1)-(stl*cfont.x_size);
	if (x==CENTER)
		x=((disp_y_size+1)-(stl*cfont.x_size))/2;
	}

	for (i=0; i<stl; i++)
		if (deg==0)
			printChar(*st++, x + (i*(cfont.x_size)), y);
		else
			rotateChar(*st++, x, y, i, deg);
}
void LCDTouch::print(String st, int x, int y, int deg)
{
	char buf[st.length()+1];
	st.toCharArray(buf, st.length()+1);
	print(buf, x, y, deg);
}
uint8_t* LCDTouch::getFont()
{
	return cfont.font;
}

uint8_t LCDTouch::getFontXsize()
{
	return cfont.x_size;
}

uint8_t LCDTouch::getFontYsize()
{
	return cfont.y_size;
}
int LCDTouch::getDisplayXSize()
{
	if (orient==PORTRAIT)
		return disp_x_size+1;
	else
		return disp_y_size+1;
}

int LCDTouch::getDisplayYSize()
{
	if (orient==PORTRAIT)
		return disp_y_size+1;
	else
		return disp_x_size+1;
}
void LCDTouch::printNumI(long num, int x, int y, int length, char filler)
{
	char buf[25];
	char st[27];
	boolean neg=false;
	int c=0, f=0;
  
	if (num==0)
	{
		if (length!=0)
		{
			for (c=0; c<(length-1); c++)
				st[c]=filler;
			st[c]=48;
			st[c+1]=0;
		}
		else
		{
			st[0]=48;
			st[1]=0;
		}
	}
	else
	{
		if (num<0)
		{
			neg=true;
			num=-num;
		}
	  
		while (num>0)
		{
			buf[c]=48+(num % 10);
			c++;
			num=(num-(num % 10))/10;
		}
		buf[c]=0;
	  
		if (neg)
		{
			st[0]=45;
		}
	  
		if (length>(c+neg))
		{
			for (int i=0; i<(length-c-neg); i++)
			{
				st[i+neg]=filler;
				f++;
			}
		}

		for (int i=0; i<c; i++)
		{
			st[i+neg+f]=buf[c-i-1];
		}
		st[c+neg+f]=0;

	}

	print(st,x,y,0);
}
void LCDTouch::printNumF(double num, byte dec, int x, int y, char divider, int length, char filler)
{
	char buf[25];
	char st[27];
	boolean neg=false;
	int c=0, f=0;
	int c2;
	long mult;
	unsigned long inum;
  
	if (dec<1)
	dec=1;
	if (dec>5)
	dec=5;
	  
	if (num==0)
	{
		if (length!=0)
		{
			for (c=0; c<(length-2-dec); c++)
				st[c]=filler;
			st[c]=48;
			st[c+1]=divider;
			for (int i=0; i<dec; i++)
				st[c+2+i]=48;
			st[c+2+dec]=0;
		}
		else
		{
			st[0]=48;
			st[1]=divider;
			for (int i=0; i<dec; i++)
				st[2+i]=48;
			st[2+dec]=0;
		}
	}
	else
	{
		if (num<0)
		{
			neg=true;
			num=-num;
		}
	  
		mult = 1;
		for (int j = 0; j < dec; j++)
			mult = mult * 10;
		inum=long(num*mult + 0.5);
	  
		while (inum>0)
		{
			buf[c]=48+(inum % 10);
			c++;
			inum=(inum-(inum % 10))/10;
		}
		if ((num<1) and (num>0))
		{
			buf[c]=48;
			c++;
		}
		while (c<(dec+1))
		{
			buf[c]=48;
			c++;
		}
		buf[c]=0;
	  
		if (neg)
		{
			st[0]=45;
		}
	  
		if (length>(c+neg-1))
		{
			for (int i=0; i<(length-c-neg-1); i++)
			{
				st[i+neg]=filler;
				f++;
			}
		}

		c2=neg;
		for (int i=0; i<c; i++)
		{
			st[c2+f]=buf[c-i-1];
			c2++;
			if ((c-(c2-neg))==dec)
			{
				st[c2+f]=divider;
				c2++;
			}
		}
		st[c2+f]=0;
	}

	print(st,x,y,0);
}
void LCDTouch::drawBitmap(int x, int y, int sx, int sy, bitmapdatatype data, int deg, int rox, int roy)
{
	unsigned int col;
	int tx, ty, newx, newy;
	byte r, g, b;
	double radian;
	radian=deg*0.0175;  

	if (deg==0)
		drawBitmap(x, y, sx, sy, data,0);
	else
	{
		for (ty=0; ty<sy; ty++)
			for (tx=0; tx<sx; tx++)
			{
				col=pgm_read_word(&data[(ty*sx)+tx]);

				newx=x+rox+(((tx-rox)*cos(radian))-((ty-roy)*sin(radian)));
				newy=y+roy+(((ty-roy)*cos(radian))+((tx-rox)*sin(radian)));

				SetXY(newx, newy, newx, newy);
				Write_Data((col<<8)|col);
			}
	}
	clrXY();
}
void LCDTouch::drawBitmap(int x, int y, int sx, int sy, bitmapdatatype data, int scale)
{
	unsigned int col;
	int tx, ty, tc, tsx, tsy;
	byte r, g, b;

	if (scale==1)
	{
		if (orient==PORTRAIT)
		{
			SetXY(x, y, x+sx-1, y+sy-1);
			for (tc=0; tc<(sx*sy); tc++)
			{
				col=pgm_read_word(&data[tc]);
				Write_Data((col>>8)|col);
			}
		}
		else
		{
			for (ty=0; ty<sy; ty++)
			{
				SetXY(x, y+ty, x+sx-1, y+ty);
				for (tx=sx; tx>=0; tx--)
				{
					col=pgm_read_word(&data[(ty*sx)+tx]);
					Write_Data((col>>8)|col);
				}
			}

		}
	}
	else
	{
		if (orient==PORTRAIT)
		{
			for (ty=0; ty<sy; ty++)
			{
				SetXY(x, y+(ty*scale), x+((sx*scale)-1), y+(ty*scale)+scale);
				for (tsy=0; tsy<scale; tsy++)
					for (tx=0; tx<sx; tx++)
					{
						col=pgm_read_word(&data[(ty*sx)+tx]);
						for (tsx=0; tsx<scale; tsx++)
						Write_Data((col>>8)|col);
					}
			}
		}
		else
		{
			for (ty=0; ty<sy; ty++)
			{
				for (tsy=0; tsy<scale; tsy++)
				{
					SetXY(x, y+(ty*scale)+tsy, x+((sx*scale)-1), y+(ty*scale)+tsy);
					for (tx=sx; tx>=0; tx--)
					{
						col=pgm_read_word(&data[(ty*sx)+tx]);
						Write_Data((col>>8)|col);
					}
				}
		}
	}
	clrXY();
	}
}