#include "Arduino.h"
#include <inttypes.h>
#include "HW_AVR_defines.h"

#define LCD2_4	1
#define LCD2_8	2
#define LCD3_2	3
#define LEFT 0

#define PREC_LOW			1
#define PREC_MEDIUM			2
#define PREC_HI				3
#define PREC_EXTREME		411

#define LATCHED_16		17
#define SERIAL_4PIN		4
#define PORTRAIT 		0
#define LANDSCAPE 		1
#define RIGHT 			9999
#define CENTER 			9998

/******************************************/
#include <SdFat.h>
#include <SD.h>
//Sd2Card card;

#define X_CONST 320
#define Y_CONST 240

#define PREC_TOUCH_CONST 10

#define PixSizeX	13.78
#define PixOffsX	411

#define PixSizeY	11.01
#define PixOffsY	378

#define WINDOW_XADDR_START	0x0050 // Horizontal Start Address Set
#define WINDOW_XADDR_END	0x0051 // Horizontal End Address Set
#define WINDOW_YADDR_START	0x0052 // Vertical Start Address Set
#define WINDOW_YADDR_END	0x0053 // Vertical End Address Set
#define GRAM_XADDR		    0x0020 // GRAM Horizontal Address Set
#define GRAM_YADDR		    0x0021 // GRAM Vertical Address Set
#define GRAMWR 			    0x0022 // memory write

#define CAL_X 				0x00378F66UL
#define CAL_Y 				0x03C34155UL
#define CAL_S 				0x000EF13FUL

#define White          0xFFFF
#define Black          0x0000
#define Blue           0x001F
#define Blue2          0x051F
#define Red            0xF800
#define Magenta        0xF81F
#define Green          0x07E0
#define Cyan           0x7FFF
#define Yellow         0xFFE0

struct _current_font
{
	uint8_t* font;
	uint8_t x_size;
	uint8_t y_size;
	uint8_t offset;
	uint8_t numchars;
};

class LCDTouch
{
	public:
		int TP_X,TP_Y;
		void LCDUTFT(byte model);
		void Write_Command(unsigned int c);
		void Write_Data(unsigned int c);
		void LCD_Write_DATA(char VH,char VL);
		void Write_Command_Data(unsigned int cmd,unsigned int dat);
		void Lcd_Init();
		void SetXY(unsigned int x0,unsigned int y0,unsigned int x1,unsigned int y1);
		void Pant(unsigned int color);
		void fillScr(unsigned int color);
		void fillScr(byte r, byte g, byte b);
		void clrScr();
		void Touch_Init(void);
		void Touch_WriteData(unsigned char data); 
		unsigned int Touch_ReadData();
		void Touch_Read();
		char Touch_DataAvailable();
		int Touch_GetX();
		int Touch_GetY();
		void setPrecision(byte precision);
		void LcdOff();
		void LcdOn();
		void clrXY();
		void drawRect(int x1, int y1, int x2, int y2);
		void drawHLine(int x, int y, int l);
		void drawVLine(int x, int y, int l);
		void drawLine(int x1, int y1, int x2, int y2);
		void setColor(word color);
		void setColor(byte r, byte g, byte b);
		void setBackColor(word color);
		void setBackColor(byte r, byte g, byte b);
		void drawRoundRect(int x1, int y1, int x2, int y2);
		void drawPixel(int x, int y);
		void setPixel(word color);
		void fillRect(int x1, int y1, int x2, int y2);
		void fillRoundRect(int x1, int y1, int x2, int y2);
		void drawCircle(int x, int y, int radius);
		void fillCircle(int x, int y, int radius);
		void printChar(byte c, int x, int y);
		void setFont(uint8_t mxsize,uint8_t mysize,uint8_t moffset);
		void rotateChar(byte c, int x, int y, int pos, int deg);
		void print(char *st, int x, int y, int deg);
		void print(String st, int x, int y, int deg);
		uint8_t* getFont();
		uint8_t getFontXsize();
		uint8_t getFontYsize();
		int getDisplayXSize();
		int getDisplayYSize();
		void printNumI(long num, int x, int y, int length, char filler);
		void printNumF(double num, byte dec, int x, int y, char divider, int length, char filler);
		void drawBitmap(int x, int y, int sx, int sy, bitmapdatatype data, int deg, int rox, int roy);
		void drawBitmap(int x, int y, int sx, int sy, bitmapdatatype data, int scale);
	protected: 
		long	_default_orientation;
		long	touch_x_left, touch_x_right, touch_y_top, touch_y_bottom;
		byte	prec;
		byte orient;
		byte fch, fcl, bch, bcl;
		long disp_x_size, disp_y_size;
		byte display_model,display_transfer_mode,display_serial_mode;
		regtype *P_RS, *P_WR, *P_CS, *P_RST, *P_SDA, *P_SCL, *P_ALE;
		regsize B_RS, B_WR, B_CS, B_RST, B_SDA, B_SCL, B_ALE;
		
		void _hw_special_init();
		void LCD_Writ_Bus(char VH,char VL, byte mode);
		void _set_direction_registers(byte mode);
		void _fast_fill_16(int ch, int cl, long pix);
		void _fast_fill_8(int ch, long pix);
		_current_font	cfont;
};

extern LCDTouch Touchtest;
